# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

from pystark import Stark
from pystark.config import ENV
from pystark.plugins.helpers import db
from pyrogram.errors import PeerIdInvalid


@Stark.cmd("botban", owner_only=True)
async def ban_users(bot: Stark, msg):
    if len(msg.command) == 1:
        user_id = None
        if msg.reply_to_message:
            for word in msg.reply_to_message.text.split():
                if word.isdigit() and len(word) == 10:
                    user_id = int(word)
        else:
            await msg.reply("Pass an ID", quote=True)
            return
        if user_id:
            reason = await bot.get_messages(msg.chat.id, msg.reply_to_message.message_id)
            try:
                reason = reason.reply_to_message.link
            except AttributeError:
                reason = None
            await db.set("bans", user_id, {"reason": reason})
            await msg.reply(f"Banned `{user_id}`")
    else:
        user_id = msg.command[1]
        if user_id.isdigit() and len(user_id) == 10:
            user_id = int(user_id)
        else:
            await msg.reply("Wrong ID", quote=True)
            return
        if user_id:
            if user_id in ENV().OWNER_ID:
                await msg.reply(f"Banned `{user_id}`", quote=True)
                await msg.reply(f"Seriously banned.", quote=True)
                await msg.reply(f"I'm not kidding.", quote=True)
                return
            await db.set("bans", user_id, {"reason": None})
            await msg.reply(f"Banned `{user_id}`", quote=True)


@Stark.cmd("botunban", owner_only=True)
async def unban_users(_, msg):
    if len(msg.command) == 1:
        await msg.reply("Please pass a user id or username", quote=True)
    else:
        user_id = msg.command[1]
        if user_id.isdigit() and len(user_id) == 10:
            user_id = int(user_id)
        else:
            await msg.reply("Wrong ID", quote=True)
            return
        if user_id:
            if user_id in await db.get_all_data("bans"):
                await db.delete("bans", user_id)
                await msg.reply(f"Unbanned `{user_id}`", quote=True)
            else:
                await msg.reply("Wasn't banned...", quote=True)


@Stark.cmd("botbanlist", owner_only=True)
async def ban_list(bot: Stark, msg):
    number = 0
    banned = await db.get_all_data("bans")
    if len(banned) == 0:
        text = "No one is banned currently"
    else:
        text = f'**Banned Users [{len(banned)}]**'
        try:
            users = await bot.get_users([i["user_id"] for i in banned])
            for user in users:
                number = number+1
                user_text = f"\n\n{number}) {user.mention} [`{user.id}`]"
                reason = [i["reason"] for i in banned if i["user_id"] == user.id]
                if reason[0]:
                    user_text += f" - Reason - {reason[0]}"
                text += user_text
        except PeerIdInvalid:
            for i in banned:
                number = number+1
                try:
                    user = await bot.get_users(i["user_id"])
                    user_text = f"\n\n{number}) {user.mention} [`{user.id}`]"
                except PeerIdInvalid:
                    user_text = f"\n\n{number}) {i['user_id']}"
                if i["reason"]:
                    user_text += f" - Reason - {i['reason']}"
                text += user_text
    await msg.reply(text, quote=True)



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

import os
import asyncio
from ..helpers import patch
from pystark.helpers import l10n
from typing import Optional, Union
from pyrogram.types import Message, User
from pyrogram.errors import MessageTooLong


@patch(Message)
class Message:
    """Message class instances represents messages. This class adds some additional methods to existing pyrogram methods of Message"""

    async def react(self, text: str, del_in: int = 0, **kwargs) -> "Message":
        """React/Reply to a message. Similar to pyrogram's `msg.reply` but takes care of:

        - Long messages - Sends long messages which raise `MESSAGE_TOO_LONG` as documents automatically.
        - Localization - For example, if your `en.yml` file has a HAS_BEEN` key then `message.react("HAS_BEEN")` will be replaced by it.
        - `del_in` argument - Extra argument to delete a message after n seconds asynchronously. If n is 0, message is not deleted at all.
        """
        text = str(text)
        lang = await l10n()
        if lang and text in lang:
            text = lang[text]
        try:
            reply = await self.reply(text, quote=True, disable_web_page_preview=True, **kwargs)
        except MessageTooLong:
            reply = await self.reply("Sending as document...", quote=True)
            file = f'{reply.message_id}.txt'
            with open(file, 'w+') as f:
                f.write(text)
            await reply.delete()
            reply = await self.reply_document(file, caption="Output")
            os.remove(file)
        if del_in:
            await asyncio.sleep(del_in)
            await reply.delete()
        return reply

    @property
    def args(self, split: str = " ") -> list[str]:
        """List arguments passed in a message. Removes first word (the command itself)

        Parameters:

            split (str, optional): Define how to split the arguments, defaults to whitespace.

        Example:

            If text is `/start reply user`, return value would be `["reply", "user"]`
        """
        args = self.text.split(split)
        args.pop(0)
        return args

    @property
    def input(self) -> str:
        """Input passed in a message. Removes first word (the command itself)

        Example:

            If text is `/start reply user`, return value would be `reply user`
        """
        return self.text.split(" ", 1)[1]

    @property
    def ref(self) -> Union[int, str, None]:
        """Returns the referred user's id or username. To get the full user, use method `get_ref_user`

        Useful to get the referent user's id or username for a command.
        If command was replied to a message, the replied message's user id is returned.
        Otherwise, the first argument of command is returned (which isn't guaranteed to be an actual user or an id).

        **Example**:

            If command `/ban` was replied to a user's message, user id (`message.reply_to_message.from_user.id`) is returned.

            If command `/ban 12345678` was sent, user id (integer) `12345678` is returned.

            If command `/ban StarkProgrammer` was sent, username (string) `StarkProgrammer` is returned.

            If command `/ban 87654321 12345678` was sent, user id (integer) `87654321` is returned but `12345678` is ignored because it's not the first argument.

            If command `/ban` was sent with no arguments and was also not a reply to any message, None is returned.

            If command `/ban ok` was sent, "ok" is returned which isn't an actual user but will not raise exception.
        """
        if self.reply_to_message:
            return self.reply_to_message.from_user.id
        args = self.args
        if not args:
            return
        return args[0] if not args[0].isdigit() else int(args[0])

    async def get_ref_user(self) -> Optional[User]:
        """Returns the full referred user. To get only user id or username, use property `ref` as it's faster.

        Useful to get the referent of a command.
        If command was replied to a message, the replied message's user is returned.
        Otherwise, the first argument of command is used to get the user.

        !!! note

            If command is not replied then first argument of command is considered user id or username. It's on you to handle PeerIdInvalid and UsernameInvalid

        **Example**:

            If command `/ban` was replied to a user's message, user (`message.reply_to_message.from_user`) is returned.

            If command `/ban 12345678` was sent, the user instance of user with id `12345678` is returned.

            If command `/ban StarkProgrammer` was sent, the user instance of user with username `StarkProgrammer` is returned.

            If command `/ban 87654321 12345678` was sent, the user instance of user with id `87654321` is returned but `12345678` is ignored because it's not the first argument.

            If command `/ban` was sent with no arguments and was also not a reply to any message, None is returned.
        """
        if self.reply_to_message:
            return self.reply_to_message.from_user
        args = self.args
        if not args:
            return
        user = args[0] if not args[0].isdigit() else int(args[0])
        user = await self._client.get_users(user)
        return user



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


from pystark.database.sql import Base
from sqlalchemy import Column, BigInteger, String


class Bans(Base):
    __tablename__ = "bans"
    __table_args__ = {'extend_existing': True}
    user_id = Column(BigInteger, primary_key=True)
    reason = Column(String, nullable=True)

    def __init__(self, user_id, reason=None):
        self.user_id = user_id
        self.reason = reason



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


# Default Configuration for your project


# Path of the 'plugins' directory in relation to the root directory or List of multiple paths
# If name of your directory is 'files' and it is in the same folder as 'bot.py', pass plugin='files'.
# Defaults to 'plugins', i.e, a folder named 'plugins' in same directory as 'bot.py'
PLUGINS = "plugins" or ["plugins"]

# Set bot menu using command description if provided
SET_BOT_MENU = True

# Prefixes for commands. For multiple prefixes, specify multiple together like ["&", "*", "/", "."]
CMD_PREFIXES = ["/"]

# The addons (default plugins) you want to load from pystark
ADDONS = [
    "start",  # /start command which replies with text in START variable and START_IN_GROUPS variables for groups.
    "help",  # /help command which replies with text in HELP variable. Only private chats
    "about",  # /about command which replies with text in ABOUT variable. Only private chats
    "id",  # /id command which replies with user id and chat id.
    # "sudo",  # Use /sudo to show all sudo commands.
    # "json",  # /json command to get Message JSON and /jsondoc command to get Message JSON as document. Only owner can use them.
    # "must_join",  # Put usernames or chat ids where users must join in MUST_JOIN variable.
    # "bans",  # Use /botban to ban people from using the bot. /botunban to unban. /botbanlist to see banned users. Needs postgresql database with table "bans"
    # "broadcast",  # Use /broadcast to broadcast a message to users. Needs postgresql database with table "users"
    # "stats",  # Use /stats to get current user stats. Needs postgresql database with table "users"
]

# Start Message. Message sent at /start command. Only needed if `start` addon is enabled
START = "Thank you for starting this bot."

# Help Message. Message sent at /help command. Only needed if `help` addon is enabled. {commands} is automatically replaced with commands if used.
HELP = """
Available Commands are:

{commands}
"""

# About Message. Message sent at /about command. Only needed if `about` addon is enabled
ABOUT = "About This Bot \nDeveloped using @PyStark"

# List of usernames or chat ids where users must join
MUST_JOIN = []

# Specify timezone for logging. Defaults to India (Asia/Kolkata)
TIMEZONE = "Asia/Kolkata"

# Only if postgresql database is used
DATABASE_TABLES = [
    # "users",  # Needed for "stats" addon and "broadcast" addon
    # "bans",  # Needed for "bans" addon
]

# Directory with localization files. This is useful:
#   - if you want your bot to be in multiple languages
#                      or
#   - if you want to keep strings separately.
# Files should have `yaml`, `yml` or `json` extension and should be in YAML or JSON Notations.
# If "english" or "en" file exists in localization directory, it's considered default. Otherwise, the first file alphabetically.
# This is optimized to not slow the bot.
LOCALIZATION = "" or None

# Start Message for groups. Leave Empty ("") to remove.
START_IN_GROUPS = "Hey :) Ping me privately if you have any questions on how to use me."



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

import sqlalchemy
from ..config import ENV
from sqlalchemy import inspect
from pystark.logger import logger
from typing import Optional, Union
from pystark.config import settings
from sqlalchemy import create_engine
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session


# Needs to be outside the class so that models can exist without database
Base = declarative_base()
__checked__ = False


class Database:
    """Class Database with in-built functions to make sqlalchemy easier.

    Attributes:
        url - DATABASE_URL
        base - SQLAlchemy Base
        session - SQLAlchemy Session
        engine - SQLAlchemy engine
    """

    def __init__(self):
        self.url = ENV.DATABASE_URL
        if not self.url:
            logger.critical('No DATABASE_URL defined. Exiting...')
            raise SystemExit
        module = settings()
        self.base = Base
        engine = create_engine(ENV.DATABASE_URL)
        Base.metadata.bind = engine
        global __checked__
        if not __checked__:
            specified_tables = module.DATABASE_TABLES
            actual_tables = Base.metadata.tables
            all_tables = ["bans", "users"]
            for t in all_tables:
                if t not in specified_tables and t in actual_tables:
                    Base.metadata.remove(actual_tables[t])
            __checked__ = True
        Base.metadata.create_all(bind=engine, checkfirst=True)
        Session = scoped_session(sessionmaker(bind=engine, autoflush=False))
        # Type-Hinting is wrong here, but a temporary way to get Hints for Session object as their methods are similar
        self.session: sqlalchemy.orm.Session = Session
        self.engine = Session.get_bind()

    async def get(self, table_name: str, primary_key, key: Union[str] = "") -> Union[dict, None]:
        """Get data from postgres database using table name as string.

        Parameters:
            table_name (``str``):
                The table name to query on.

            primary_key:
                The value of a primary_key to query the table.

            key:
                If you only need one key.

        Returns:
            Key-value pairs if no key is passed. Value of the key if key is passed. None if no such row.

        Raises:
            TableNotFound- If table is not found
        """
        tables_dict = await self.tables_dict()
        if table_name not in tables_dict:
            raise TableNotFound(f'No such table exists : {table_name}')
        table = tables_dict[table_name]
        query = self.session.query(table).get(primary_key)
        if not query:
            return
        all_val = await self._class_vars(query)
        if key:
            return all_val.get(key)
        return all_val

    async def set(self, table_name: str, primary_key, data: dict = None):
        """Set data in postgres database using table name as string. ``insert`` is an alias of ``set``

        If `data` is not passed, `primary key` is added in a new row to table.

        Parameters:
            table_name (``str``):
                The table name to query on.

            primary_key:
                The value of a primary_key to query the table.

            data (``dict``):
                Key-value pairs of column name and row value as dictionary

        Returns:
            ``bool``: True on success
        """
        if data is None:
            data = {}
        tables_dict = await self.tables_dict()
        if table_name not in tables_dict:
            raise TableNotFound(f'No such table exists : {table_name}')
        table = tables_dict[table_name]
        try:
            query = self.session.query(table).get(primary_key)
            if query:
                for i in data:
                    setattr(query, i, data[i])
            else:
                row = table(primary_key)
                for i in data:
                    setattr(query, i, data[i])
                self.session.add(row)
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

    async def count(self, table_name: str) -> Optional[int]:
        """Get number of rows in postgres table.

        Parameters:
            table_name:
                The table name to query on.
        """
        tables_dict = await self.tables_dict()
        if table_name not in tables_dict:
            raise TableNotFound(f'No such table exists : {table_name}')
        table = tables_dict[table_name]
        count = self.session.query(table).count()
        return count

    insert = set

    async def all(self, table_name: str) -> Optional[list[dict]]:
        """Get all rows in postgres table. Better use ``get_all_data``

        Parameters:
            table_name (``str``):
                The table name to query on.
        """
        tables_dict = await self.tables_dict()
        if table_name not in tables_dict:
            raise TableNotFound(f'No such table exists : {table_name}')
        table = tables_dict[table_name]
        try:
            try:
                all_ = self.session.query(table).all()
                self.session.close()
                return [await self._class_vars(d) for d in all_]
            except ProgrammingError:
                self.session.rollback()
                data = []
                raw = self.session.execute(f'SELECT * FROM {table_name}')
                for r in raw:
                    row = {}
                    for key, value in r.items():
                        row.update({key: value})
                    data.append(row)
                return data
        except Exception:
            self.session.rollback()
            raise

    async def delete(self, table_name: str, primary_key) -> None:
        """Delete row in postgres database using table name as string.

        Parameters:
            table_name (``str``):
                The table name to query on.

            primary_key:
                The value of a primary_key to query the table.
        """
        tables_dict = await self.tables_dict()
        if table_name not in tables_dict:
            raise TableNotFound(f'No such table exists : {table_name}')
        table = tables_dict[table_name]
        try:
            query = self.session.query(table).get(primary_key)
            self.session.delete(query)
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

    async def get_all_primary_keys(self, table_name) -> list:
        """Get all primary keys of the table

        Parameters:
            table_name (``str``):
                Name of the table whose primary keys you need.
        """
        return [i[0] for i in self.engine.execute('SELECT * FROM %s' % table_name)]

    async def get_all_data(self, table_name: str) -> list[dict]:
        """Get all data of the table as a list of rows as dictionaries

        Parameters:
            table_name (``str``):
                Name of the table whose data you need.
        """
        return [getattr(x, "_asdict")() for x in self.engine.execute('SELECT * FROM %s' % table_name)]

    async def change_column_type(self, table_name: str, column_name: str, column_type: str):
        """Change Column Type of Table using raw sql queries. Supposed to be run only one time for any particular values.

        Parameters:
            table_name (``str``):
                Name of the table where the column exists.

            column_name:
                Name of the column you want to change.

            column_type:
                New Column Type.
        """
        self.engine.execute('ALTER TABLE %s ALTER COLUMN %s TYPE %s' % (table_name, column_name, column_type))

    async def add_column(self, table_name: str, column_name: str, column_type: str):
        """Add Column of Table using raw sql queries. Supposed to be run only one time for any particular values.

        Parameters:
            table_name (``str``):
                Name of the table where you want the table.

            column_name:
                Name of the column you want to add.

            column_type:
                New Column Type.
        """
        self.engine.execute('ALTER TABLE %s ADD %s %s' % (table_name, column_name, column_type))

    async def remove_column(self, table_name: str, column_name: str):
        """Remove Column of Table using raw sql queries. Supposed to be run only one time for any particular values.

        Parameters:
            table_name (``str``):
                Name of the table where the table exists.

            column_name:
                Name of the column you want to delete.
        """
        self.engine.execute('ALTER TABLE %s DROP COLUMN %s' % (table_name, column_name))

    async def columns(self, table_name: str) -> list[tuple[str]]:
        """List of all public column names in the database

        Parameters:
            table_name (``str``):
                Name of the table whose columns you need.
        """
        return [x[0] for x in self.engine.execute("SELECT column_name FROM information_schema.columns WHERE table_schema = 'public' AND table_name = '%s'" % table_name)]

    async def list_tables(self) -> list:
        """Returns list all tables in your database"""
        return inspect(self.engine).get_table_names()

    async def has_table(self, table_name: str) -> bool:
        """Returns True if table exists else False. To also check if it's sqlalchemy model exists, use function ``table_exists`` instead"""
        inspect(self.engine)
        return inspect(self.engine).has_table(table_name)

    async def table_exists(self, table_name: str) -> bool:
        """Returns True if table exists with sqlalchemy model else False. To check only for table, use function ``has_table`` instead"""
        tables_dict = await self.tables_dict()
        if table_name in tables_dict:
            return True
        else:
            return False

    async def tables_dict(self) -> dict:
        """Returns all {tablename: table} where tablename is a string and table is a model (python class)"""
        return {table.__tablename__: table for table in self.base.__subclasses__()}

    @staticmethod
    async def _class_vars(class_obj) -> dict:
        v = vars(class_obj)
        return {key: v[key] for key in v if key != "_sa_instance_state"}


class TableNotFound(Exception):
    pass



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


from pystark.database.sql import Base
from sqlalchemy import Column, BigInteger, String


class Users(Base):
    __tablename__ = "users"
    __table_args__ = {'extend_existing': True}
    user_id = Column(BigInteger, primary_key=True)
    lang = Column(String)

    def __init__(self, user_id: int, lang: str = "en"):
        self.user_id = user_id
        self.lang = lang



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


import os
from typing import Union
from .logger import logger
from dotenv import load_dotenv
from pystark import settings as s
from importlib import import_module
from pystark.plugins import stark as h


load_dotenv('.env')

default = None


def settings():
    global default
    if default:
        return default
    if os.path.exists('settings.py'):
        mod = import_module('settings')
    elif os.path.exists('data.py'):
        logger.warn("File name 'data.py' for settings is deprecated. Please rename it to 'settings.py' instead.")
        mod = import_module('data')
    else:
        logger.warn("Settings file not found. Default values will be used.")
        mod = import_module('pystark.settings')
    for i in dir(s):
        if not i.isupper():
            continue
        if "STARKBOTS" in mod.__dict__ and getattr(mod, "STARKBOTS") and i == "ADDONS":
            continue
        try:
            getattr(mod, i)
        except AttributeError:
            setattr(mod, i, getattr(s, i))
    for i in ["STARKBOTS", "FULL_MESSAGES", "CUSTOM_USERS_TABLE", "NO_BUTTONS", "REMOVE_ADDONS"]:  # my personalization (not in settings.py)
        try:
            getattr(mod, i)
        except AttributeError:
            setattr(mod, i, False)
    prefixes = getattr(mod, "CMD_PREFIXES")
    if not isinstance(prefixes, list):
        setattr(mod, "CMD_PREFIXES", [prefixes])
    if getattr(mod, "STARKBOTS"):

        # FULL_MESSAGES
        if "FULL_MESSAGES" not in mod.__dict__ or not getattr(mod, "FULL_MESSAGES"):
            # START
            setattr(mod, "START", h.START.replace("{1}", getattr(mod, "START")))
            # HELP
            setattr(mod, "HELP", h.HELP.replace("{1}", getattr(mod, "HELP")))
            # REPO
            try:
                rep = getattr(mod, "REPO")
            except AttributeError:
                rep = ""
            # ABOUT
            setattr(
                mod,
                "ABOUT",
                h.ABOUT
                .replace("{1}", getattr(mod, "ABOUT"))
                .replace("{2}", (("\n\nSource Code : [Click Here](https://github.com/StarkBotsIndustries/"+rep+")") if rep else ""))
            )
        # MUST_JOIN
        setattr(mod, "MUST_JOIN", ["StarkBots", "StarkBotsChat"])

        # ADDONS
        if "ADDONS" not in mod.__dict__:
            from pystark.plugins import addons
            plugs = [f for f in os.listdir(addons.__path__[0]) if f.endswith(".py")]
            if getattr(mod, "NO_BUTTONS"):
                plugs.remove("buttons.py")
            rem = getattr(mod, "REMOVE_ADDONS")
            if rem:  # Bad code?
                for p in rem:
                    if p in plugs:
                        plugs.remove(p)
                    elif p+".py" in plugs:
                        plugs.remove(p+".py")
            setattr(mod, "ADDONS", plugs)

        # DATABASE_TABLES
        tables = getattr(mod, "DATABASE_TABLES")
        if "bans" not in tables:
            tables.append("bans")

        # CUSTOM_USERS_TABLE
        if ("CUSTOM_USERS_TABLE" not in mod.__dict__ or not getattr(mod, "CUSTOM_USERS_TABLE")) and "users" not in tables:
            tables.append("users")
        setattr(mod, "DATABASE_TABLES", tables)

        # CMD_PREFIXES
        setattr(mod, "CMD_PREFIXES", ["/", "!"])

    # Set Timezone to env so there's no circular import for logger.py
    os.environ["TIMEZONE"] = getattr(mod, 'TIMEZONE')
    default = mod  # Cache
    return mod


class ENV:
    API_ID = os.environ.get("API_ID", "").strip()
    API_HASH = os.environ.get("API_HASH", "").strip()
    BOT_TOKEN = os.environ.get("BOT_TOKEN", "").strip()
    OWNER_ID = os.environ.get("OWNER_ID", "").strip().split(" ")
    LOG_CHAT = os.environ.get("LOG_CHAT", "").strip()
    SUDO_USERS = os.environ.get("SUDO_USERS", "").strip().split(" ")
    DB_SESSION = os.environ.get("DB_SESSION", "").strip()
    DB_CHAT_ID = os.environ.get("DB_CHAT_ID", "").strip()
    DATABASE_URL = os.environ.get("DATABASE_URL", "").strip()
    if DATABASE_URL:
        if 'postgres' in DATABASE_URL and 'postgresql' not in DATABASE_URL:
            DATABASE_URL = DATABASE_URL.replace("postgres", "postgresql")
    REDIS_URL = os.environ.get("REDIS_URL", "").strip()
    REDIS_PASSWORD = os.environ.get("REDIS_PASSWORD", "").strip()

    def __init__(self):
        module = settings()

        if not self.API_ID:
            logger.critical("No API_ID found. Exiting...")
            raise SystemExit
        if not self.API_HASH:
            logger.critical("No API_HASH found. Exiting...")
            raise SystemExit
        if not self.BOT_TOKEN:
            logger.critical("No BOT_TOKEN found. Exiting...")
            raise SystemExit
        if not self.LOG_CHAT:
            self.LOG_CHAT = 0

        if "special" in self.OWNER_ID or getattr(module, "STARKBOTS"):
            if not self.OWNER_ID:
                self.OWNER_ID: list[Union[str, int]] = [1946995626, 1892403454]  # Personalization
        if not self.OWNER_ID or not self.OWNER_ID[0]:
            OWNER_ID = [0]
            # logger.warn("No OWNER_ID found. Please set one. Exiting...")
            # raise SystemExit

        try:
            self.API_ID = int(self.API_ID)
        except ValueError:
            logger.critical("API_ID is not a valid integer. Exiting...")
            raise SystemExit
        if self.DB_CHAT_ID:
            try:
                self.DB_CHAT_ID = int(self.DB_CHAT_ID)
            except ValueError:
                logger.critical("DB_CHAT_ID is not a valid integer. Exiting...")
                raise SystemExit
        owners = []
        for o in self.OWNER_ID:
            if isinstance(o, str) and o.isdigit():
                owners.append(int(o))
            else:
                owners.append(o)
        self.OWNER_ID = owners

        sudos = []
        for o in self.SUDO_USERS:
            if o.isdigit():
                sudos.append(int(o))
            else:
                sudos.append(o)
        self.SUDO_USERS = sudos

        if self.LOG_CHAT and "special" in self.LOG_CHAT and getattr(module, "STARKBOTS"):
            self.LOG_CHAT = -1001567003949  # Stark Bots Logs
        try:
            self.LOG_CHAT = int(self.LOG_CHAT)
        except ValueError:
            logger.critical("LOG_CHAT is not a valid integer. Exiting...")
            raise SystemExit



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


from typing import Union
from ..config import settings, ENV
from pyrogram import filters as f
from pyrogram.methods.decorators.on_message import OnMessage


command_data = {"total_commands": 0, "all_commands": {}, "sudo_commands": []}


class Command(OnMessage):
    @staticmethod
    def command(
        cmd: Union[str, list[str]] = None,
        description: str = None,
        group: int = 0,
        owner_only: bool = False,
        sudo_only: bool = False,
        private_only: bool = False,
        group_only: bool = False,
        channel_only: bool = False,
        extra_filters=None
    ):
        """This decorator is used to handle messages. Mainly used to create commands. All arguments are  optional.
        You can also use the alias ``Stark.cmd`` instead of ``Stark.command``.

        Parameters:

            cmd (str | list[str], optional): Command(s) that triggers your function. Defaults to None, which is helpful you only want to use extra_filters argument.

            description (str,  optional): Command description to create Bot Menu. Defaults to None. [Read More](/topics/bot-menu)

            group (int, optional): Define a group for this handler. Defaults to 0. [Read More](https://docs.pyrogram.org/topics/more-on-updates#handler-groups)

            owner_only (bool, optional): Allow only owner to use this command. Defaults to False.

            sudo_only (bool, optional): Allow only sudos to use this command. Includes owner as sudo automatically. Defaults to False.

            private_only (bool, optional): Only handle messages for private chats. Bot will ignore messages in groups and channels. Defaults to False.

            group_only (bool, optional): Only handle messages for groups. Bot will ignore messages in private chats and channels. Defaults to False.

            channel_only (bool, optional): Only handle messages for channels. Bot will ignore messages in private chats and groups. Defaults to False.

            extra_filters (pyrogram.filters, optional): Extra filters to apply in your function. Import ``filters`` from pyrogram or pystark to use this. See example below.

        Examples:

            ```python
            from pystark import Stark

            # The normal way. Bot will reply to command ``/greet`` sent anywhere and by anyone.
            @Stark.command('greet', description='Greet the user')

            # or
            @Stark.cmd('greet', 'Greet the user')

            # Bot will reply only to owner, that is, the user whose id is set as OWNER_ID in environment variables.
            # Others will be ignored.
            @Stark.command('greet', owner_only=True)

            # Bot will reply only to sudo users or owner, that is, users set as SUDO_USERS or OWNER_ID in environment variables.
            # Others will be ignored.
            @Stark.command('greet', sudo_only=True)

            # Bot will reply only if message is sent in private chat (aka pm).
            # Messages in groups and channels will be ignored.
            @Stark.command('greet', private_only=True)

            # Bot will reply only if message is sent in groups.
            # Messages in groups and private chats will be ignored.
            @Stark.command('greet', group_only=True)

            # Bot will reply only if message is sent in channels.
            # Messages in private chats and groups will be ignored.
            @Stark.command('greet', channel_only=True)


            # Filter all messages.

            # Use positive integer to execute after executing another function in default group that also filtered this message.
            @Stark.command(group=1)

            # or Use negative integer to execute before executing another function in default group that also filtered this message.
            @Stark.command(group=-1)

            # Don't use this as other functions won't work.
            @Stark.command()


            # Filter other type of messages using extra_filters.

            # Import filters from pyrogram or pystark.
            from pystark import filters

            # Filter only media messages.
            @Stark.command(extra_filters=filters.media)

            # Filter only text messages.
            @Stark.command(extra_filters=filters.text)

            # Filter only messages sent by 'StarkProgrammer'.
            @Stark.command(extra_filters=filters.user('StarkProgrammer'))

            # Filter only messages sent in 'StarkBotsChat'
            @Stark.command(extra_filters=filters.chat('StarkBotsChat'))

            # Filter only messages with the word 'baby'.
            @Stark.command(extra_filters=filters.regex('baby'))

            # Filter all media messages sent by bots.
            @Stark.command(extra_filters=filters.bot & filters.media)

            # Filter all messages sent by bots except media messages.
            @Stark.command(extra_filters=filters.bot & ~filters.media)

            # Filter either media messages or text messages.
            @Stark.command(extra_filters=filters.text | filters.media)
            ```
        """
        if isinstance(cmd, str):
            cmd = [cmd]
        if owner_only or sudo_only:
            command_data["sudo_commands"] += cmd
        prefixes = settings().CMD_PREFIXES
        if not cmd and not extra_filters:
            filters_ = f.all
        elif cmd and extra_filters:
            for c in cmd:
                if c not in command_data["all_commands"]:
                    command_data["total_commands"] += 1
                    command_data["all_commands"][c] = ""
            filters_ = f.command(cmd, prefixes=prefixes) & extra_filters
        elif extra_filters:
            filters_ = extra_filters
        else:
            for c in cmd:
                if c not in command_data["all_commands"]:
                    command_data["total_commands"] += 1
                    command_data["all_commands"][c] = ""
            filters_ = f.command(cmd, prefixes=prefixes)
        if cmd and description:
            for c in cmd:
                command_data["all_commands"][c] = description
        if sudo_only:
            filters_ = filters_ & f.user(ENV().SUDO_USERS+ENV().OWNER_ID)
        elif owner_only:
            filters_ = filters_ & f.user(ENV().OWNER_ID)
        if private_only:
            filters_ = filters_ & f.private
        if group_only:
            filters_ = filters_ & f.group
        if channel_only:
            filters_ = filters_ & f.channel
        decorator = OnMessage.on_message(filters_, group)
        return decorator

    cmd = command  # alias



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


aliases = {
    'edit': 'can_be_edited',
    'manage': 'can_manage_chat',
    'delete': 'can_delete_messages',
    'restrict': 'can_restrict_members',
    'promote': 'can_promote_members',
    'change': 'can_change_info',
    'invite': 'can_invite_users',
    'pin': 'can_pin_messages',
    'vc': 'can_manage_voice_chats',
}


class Admins:
    @staticmethod
    def admins(alias: str = ""):
        """This decorator is used to allow only admins to use a particular command.

        Parameters:

            alias (str, optional): Permission to check for the user. Can be `edit`/`can_be_edited` or `manage`/`can_manage_chat` or
                `delete`/`can_delete_messages` or `restrict`/`can_restrict_members` or `promote`/`can_promote_members` or `change`/`can_change_info`
                `invite`/`can_invite_users` or `pin`/`can_pin_messages` or `vc`/`can_manage_voice_chats`

        Examples:

            ```python
            from pystark import Stark


            # Bot will only execute function, if user who sent `/del` command has admin right `can_delete_messages`.
            @Stark.admins('delete')  # or @Stark.admins('can_delete_messages')
            @Stark.cmd("del")
            async def delete_message(bot, msg):
                await msg.reply_to_message.delete()


            # Bot will only execute function, if user who sent `/kick` command has admin right `can_restrict_members`.
            @Stark.admins('restrict')  # or @Stark.admins('can_restrict_members')
            @Stark.cmd("kick")
            async def kick_user(bot, msg):
                ...


            # Bot will only execute function, if user who sent `/play` command has admin right `can_manage_voice_chats`.
            @Stark.admins('vc')  # or @Stark.admins('can_manage_voice_chats')
            @Stark.cmd("play")
            async def play_song(bot, msg):
                ...
            ```
        """

        def first_decorator(func):
            if alias and alias in aliases:
                perm = aliases[alias]
            elif alias in list(aliases.values()):
                perm = alias
            else:
                perm = None

                # if not am_admin:
                #     await msg.react("I am not an admin here. That's so sad. Why am I even here?")
                # elif not info2[perm]:
                #     await msg.react(f"I am missing permission: {perm}")
                # elif not is_admin:
                #     await msg.react('You are not an admin here')
                # elif not info[perm]:
                #     await msg.react(f"You are missing permission: '{perm}'")

            async def wrapper(*args, **kwargs):
                bot = args[0]
                msg = args[1]
                bot_id = (await bot.get_me()).id
                bot_status = (await msg.chat.get_member(bot_id))
                if bot_status.status != 'administrator':
                    await msg.react("I_AM_NOT_ADMIN")
                    return
                if perm and not bot_status[perm]:
                    await msg.react("I_DONT_HAVE_RIGHT".format(perm))
                    return
                user_status = await msg.chat.get_member(msg.from_user.id)
                if user_status.status not in ['creator', 'administrator']:
                    await msg.react("USER_NOT_ADMIN")
                    return
                if perm and not user_status[perm]:
                    await msg.react("USER_DOESNT_HAVE_RIGHT".format(perm))
                    return
                if msg.reply_to_message:
                    user_id = msg.reply_to_message.from_user.id
                else:
                    args = msg.args
                    if not args:
                        await msg.react("NO_ID_PASSED")
                        return
                    user_id = args[0]
                    try:
                        user_id = (await bot.get_users(user_id)).id
                    except IndexError:
                        await msg.react("USER_NOT_FOUND")
                        return
                a_list = await bot.get_chat_members(msg.chat.id, filter="administrators")
                admin_user_ids = [a.user.id for a in a_list]
                if user_id in admin_user_ids:
                    if user_id == bot_id:
                        await msg.react("BAN_MYSELF")
                    else:
                        await msg.react("ACT_ON_ADMINS")
                    return
                await func(*args, **kwargs)

            return wrapper

        return first_decorator



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

from pystark import Stark
from pystark.plugins.stark import HOME_BUTTON, MAIN_BUTTONS
from pyrogram.types import InlineKeyboardMarkup, CallbackQuery
from pystark.plugins.helpers import replace, module, replace_commands


@Stark.callback(query=['home', 'about', 'help'])
async def basic_cb(bot: Stark, cb: CallbackQuery):
    # No Warns Cuz Personalized
    chat_id = cb.from_user.id
    message_id = cb.message.message_id
    if cb.data == 'home':
        text = await replace(module.START, cb.message, bot)
        buttons = MAIN_BUTTONS
    elif cb.data == 'about':
        text = await replace(module.ABOUT, cb.message, bot)
        buttons = HOME_BUTTON
    else:
        text = str(module.HELP)
        if "{commands}" in text:
            text = await replace_commands(bot, text)
        text = await replace(text, cb.message, bot)
        buttons = HOME_BUTTON
    await bot.edit_message_text(
        chat_id=chat_id,
        message_id=message_id,
        text=text,
        reply_markup=InlineKeyboardMarkup(buttons),
        disable_web_page_preview=True,
    )



# # PyStark - Python add-on extension to Pyrogram
# # Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
# #
# # This file is part of PyStark.
# #
# # PyStark is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# #
# # PyStark is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# # GNU General Public License for more details.
# #
# # You should have received a copy of the GNU General Public License
# # along with PyStark. If not, see <https://www.gnu.org/licenses/>.
#
#
# import httpx
#
#
# class HTTPError(Exception):
#     """Raised if HTTP Request to the url fails"""
#     pass
#
#
# class Requests:
#     """Send asynchronous or synchronous requests to any url (generally api) which returns json response using Httpx"""
#
#     def sync_get(self, url: str, params: dict = None) -> dict[str]:
#         """
#         Do a synchronous get request to any url using Httpx.
#
#         Parameters:
#
#         url (str) - The url to which you want to request.
#
#         Returns:
#
#         dict - JSON type response from the url
#
#         Raises:
#
#         HTTPError - Raised if the client is unable to interact with the API
#         """
#         return self._sync_request(url, "get")
#
#     def sync_post(self, url: str, params: dict = None) -> dict[str]:
#         """
#         Do a synchronous post request to any url using Httpx.
#
#         Parameters:
#
#         url (str) - The url to which you want to request.
#
#         Returns:
#
#         dict - JSON type response from the url
#
#         Raises:
#
#         HTTPError - Raised if the client is unable to interact with the API
#         """
#         return self._sync_request(url, "post")
#
#     @staticmethod
#     def _sync_request(url: str, method: str, params: dict = None) -> dict[str]:
#         with httpx.Client() as client:
#             if method == "get":
#                 response = client.get(url, params=params)
#             elif method == "post":
#                 response = client.post(url)
#             else:
#                 print("Only GET or POST requests are allowed")
#                 return {}
#             if not response.is_success:
#                 raise HTTPError(response)
#             return response.json()
#
#     @staticmethod
#     async def async_request(url: str) -> dict[str]:
#         """
#         Do an asynchronous request to any url using HTTPX.
#
#         Parameters:
#
#         url (str) - The url to which you want to request.
#
#         Returns:
#
#         dict - JSON type response from the url
#
#         Raises:
#
#         HTTPError - Raised if the client is unable to interact with the API
#         """
#         async with httpx.AsyncClient() as client:
#             client: httpx.AsyncClient
#             response = await client.get(url)
#             if not response.is_success:
#                 raise HTTPError(response)
#             return response.json()
#
#     @staticmethod
#     async def request(url: str) -> dict[str]:
#         """Alias of `pystark.helpers.Requests.async_request`"""
#         return await Requests.async_request(url)


# INIT

# from .requests import Requests


# Future Roadmap/Features

# ➥ admins_only decorator to allow only admins to use some commands
#
# ➥ List commands in help message.
#
# ➥ Userbot support (maybe)
#
# ➥ Log to TG in real-time with multiple facilities
#
# ➥ Special functions to execute shell commands properly
#
# ➥ Both synchronous and asynchronous http requests functions using httpx to ease things even more.
#
# ➥ Ease Inline Queries
#
# ➥ Ease buttons creation
#
# ➥ Buttons in default plugins
#
# ➥ Default inline queries
#
# ➥ MongoDB support (probably not)
#
# ➥ Paginate buttons
#
# ➥ pystark.helpers for more such useful helpers
#
# ➥ Your opinion (can be provided in @StarkBotsChat)


# mkdocs.yml
# - helpers / requests.md

# changelog
#     - Added helpers for sync and async http requests in `pystark.helpers.requests`

# docs/
# <br>
#
# - **requests** - Send http requests synchronously or asynchronously. [Read More Here](/helpers/requests)



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

from pystark import Stark, filters
from pystark.config import settings, ENV
from pyrogram.types import InlineKeyboardMarkup, InlineKeyboardButton, Message
from pyrogram.errors import ChatAdminRequired, UserNotParticipant, ChatWriteForbidden


@Stark.cmd(private_only=True, extra_filters=~filters.edited & filters.incoming, group=-1)
async def must_join_channel(bot: Stark, msg: Message):
    if ENV.DATABASE_URL:
        from pystark.plugins.helpers import db
        if (await db.has_table("bans")) and msg.from_user.id in (await db.get_all_primary_keys("bans")):
            await msg.reply("You have been banned from using this bot !", quote=True)
            await msg.stop_propagation()
            return
    module = settings()
    try:
        chats = module.MUST_JOIN
    except AttributeError:
        chats = []
    if not chats:
        Stark.log("Addon must_join is loaded but no chat is listed in MUST_JOIN var of settings", "warn")
        return
    nope = []
    for chat in chats:
        try:
            await bot.get_chat_member(chat, msg.from_user.id)
        except UserNotParticipant:
            nope.append(chat)
        except ChatAdminRequired:
            Stark.log(f"I'm not admin in the {chat} Chat !", "warning")
    if not nope:
        return
    buttons = []
    for i in nope:
        c = await bot.get_chat(i)
        if isinstance(i, int):
            i = c.title
        buttons.append([InlineKeyboardButton(f"✨ Join {i} ✨", url=c.invite_link)])
    try:
        await msg.reply(
            f"You must join my {'chats' if len(nope)>1 else 'chat'} to use me. Click on below buttons to join {'them' if len(nope)>1 else 'it'} \n\nAfter joining try again !",
            disable_web_page_preview=True,
            reply_markup=InlineKeyboardMarkup(buttons)
        )
        await msg.stop_propagation()
    except ChatWriteForbidden:
        pass



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


import os
import sys
import argparse
from pystark.constants import __description__, __version__


def main():
    parser = argparse.ArgumentParser(
        prog='pystark',
        description=__description__,
        usage='%(prog)s [options]',
        epilog='Enjoy the program :)',
        allow_abbrev=False,
        add_help=False
    )
    cwd = os.getcwd()
    parser.add_argument('-v', '--version', help='check the current pystark version installed', action='store_true')
    parser.add_argument('-bp', '--boilerplate', help='create boilerplate in current folder', action='store_true')
    parser.add_argument('-bph', '--boilerplate-heroku', help='create boilerplate in current folder with added heroku support', action='store_true')
    parser.add_argument('-s', '--special', help=argparse.SUPPRESS, action='store_true')
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    args = parser.parse_args()
    if args.version:
        print(f'v{__version__}')
        return
    if args.boilerplate_heroku or args.boilerplate or args.special:
        if args.special:
            print('Generating Boilerplate (for Stark Bots)...')
            boilerplate(special=True)
        elif args.boilerplate_heroku:
            print('Generating Boilerplate (with Heroku Support)...')
            boilerplate()
        else:
            print('Generating Boilerplate...')
            boilerplate()
            files = ['app.json', 'Procfile', 'README.md', 'requirements.txt']
            for file in files:
                os.remove(cwd+'/boilerplate/'+file)
        print('Done. Boilerplate is ready!')


def boilerplate(special=False):
    os.system('pip install github-clone --quiet')
    if special:
        os.system('ghclone https://github.com/StarkBotsIndustries/PyStark/tree/master/boilerplate_starkbots')
    else:
        os.system('ghclone https://github.com/StarkBotsIndustries/PyStark/tree/master/boilerplate')
    os.system('pip uninstall github-clone --quiet -y')



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


import asyncio
import subprocess


def exec_sync(cmd: str, shell: bool = False) -> (str, str):
    """Execute a system command synchronously using Python and get the stdout and stderr as strings.

    Parameters:
        cmd: Command to execute.
        shell: Whether to run in shell mode.

    Returns:
        tuple of stdout and stderr as strings
    """
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=shell)
    stdout, stderr = proc.communicate()
    return str(stdout), str(stderr)


async def exec_async(cmd: str, shell: bool = False) -> (str, str):
    """Execute a system command asynchronously using Python and get the stdout and stderr as strings

    Parameters:
        cmd: Command to execute.
        shell: Whether to run in shell mode.

    Returns:
        tuple of stdout and stderr as strings
    """
    if shell:
        proc = await asyncio.create_subprocess_shell(cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT, shell=True)
    else:
        proc = await asyncio.create_subprocess_shell(cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT, shell=True)
    stdout, stderr = await proc.communicate()
    return str(stdout), str(stderr)



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.

from pystark import Stark
try:
    from telegramdb import TelegramDB, DataPack, Member
except ImportError:
    import os
    Stark.log('Installing TelegramDB...')
    os.system('pip3 install TelegramDB==0.2.0')
    from telegramdb import TelegramDB, DataPack, Member
import os
import sys
import traceback
from ..config import ENV
from pyrogram import Client
from ..logger import logger

env = ENV()
DB_SESSION = env.DB_SESSION
DB_CHAT_ID = env.DB_CHAT_ID
API_ID = env.API_ID
API_HASH = env.API_HASH

if not DB_SESSION:
    Stark.log('No DB_SESSION defined. Exiting...', "critical")
    raise SystemExit
userbot = Client(DB_SESSION, api_id=API_ID, api_hash=API_HASH)
userbot.start()

if not DB_CHAT_ID:
    Stark.log("No DB_CHAT_ID found. Creating a chat for database", "warning")
    channel = userbot.create_channel("PyStark Telegram DB")
    Stark.log(f"Telegram DB Channel ID is {channel.id}")
    DB_CHAT_ID = channel.id

if isinstance(DB_CHAT_ID, str) and DB_CHAT_ID[1:].isdigit():
    DB_CHAT_ID = int(DB_CHAT_ID)


sys.stdout = open(os.devnull, 'w')
try:
    Session = TelegramDB(userbot, DB_CHAT_ID, debug=True, logger=logger)
    sys.stdout = sys.__stdout__
except Exception as e:
    sys.stdout = sys.__stdout__
    Stark.log(str(e), "critical")
    Stark.log("Your DB_CHAT_ID is incorrect.", "critical")
    Stark.log(traceback.format_exc(), "critical")
    raise SystemExit
Stark.log("Initializing TelegramDB [Copyright (C) 2022 anonyindian]")



# PyStark - Python add-on extension to Pyrogram
# Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries>
#
# This file is part of PyStark.
#
# PyStark is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyStark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyStark. If not, see <https://www.gnu.org/licenses/>.


import os
import struct
import asyncio
import logging
import importlib
import importlib.util
from typing import Union
from pystark.logger import logger
from pyrogram.types import BotCommand
from pyrogram import Client, idle, raw
from pystark.decorators import Mechanism
from pystark.config import ENV, settings
from pystark.constants import __version__
from inspect import getmembers, isfunction
from pystark.plugins.models.bans import Bans
from pystark.plugins.models.users import Users
from pystark.helpers import l10n_setup, get_all_langs
from pystark.decorators.command import command_data
from pyrogram.errors import ApiIdInvalid, AccessTokenInvalid, AuthKeyDuplicated, AuthKeyUnregistered, UserDeactivated


__data__ = {"total_plugins": 0, "all_plugins": {}}
__printed__ = False


class Stark(Client, Mechanism):
    __support__ = "StarkBotsChat"
    __updates__ = "pystark"
    __channel__ = "StarkBots"

    # __id__ = None
    # __username__ = None

    def __init__(self, **kwargs):
        global __printed__
        if not __printed__:
            print(f'Welcome to PyStark [v{__version__}]')
            print('Copyright (C) 2021-2022 Stark Bots <https://github.com/StarkBotsIndustries> \n')
            __printed__ = True
        env = ENV()
        super().__init__(
            ":memory:",
            api_id=env.API_ID,
            api_hash=env.API_HASH,
            bot_token=env.BOT_TOKEN,
            **kwargs
        )

    def activate(self):
        """Main method of `Stark` class which loads plugins and activates/runs your bot."""
        module = settings()
        try:
            self.log("Starting the Bot")
            self.start()
            self.log("Loading Modules")
            plugins = getattr(module, "PLUGINS")
            if plugins:
                if isinstance(plugins, list):
                    for folder in plugins:
                        Stark.log(f"Searching for plugins in '{folder}'...")
                        self.load_modules(folder)
                else:
                    self.load_modules(plugins)
            addons = module.ADDONS
            if addons:
                self.log("Loading Addons")
                for i in addons:
                    path = f"pystark/plugins/addons/{i}"
                    if not path.endswith(".py"):
                        path += ".py"
                    self.load_modules(path)
            if module.DATABASE_TABLES:
                self.log("Initializing Database")
                from pystark.database.sql import Database
                Database()  # tables need to be bound to engine
                for i in module.DATABASE_TABLES:
                    self._load_sql_model(i)
            if l10n_setup():
                self.log("Localization setup completed!")
            if getattr(module, "SET_BOT_MENU"):
                self.log("Updating Bot Menu")
                self.update_bot_menu()
            else:
                self.log("Skipping Bot Menu Update")
            logger.info("{} is now running...".format('@' + self.get_me().username))
            idle()
        finally:
            if getattr(module, "SET_BOT_MENU"):
                try:
                    self.remove_bot_menu()
                except ConnectionError:
                    pass
            self.stop()
            logger.info("Bot has stopped working. For issues, visit <https://t.me/StarkBotsChat>")

    def run(self):
        """Alias of `Stark.activate()`"""
        self.activate()

    def start(self):
        try:
            super().start()
            return
        except ApiIdInvalid:
            logger.critical("API_ID and API_HASH combination is incorrect.")
        except AccessTokenInvalid:
            logger.critical("BOT_TOKEN is invalid.")
        except (AuthKeyUnregistered, AuthKeyDuplicated, struct.error):
            logger.critical("Your STRING_SESSION is invalid. Please terminate it and generate a new one.")
        except UserDeactivated:
            logger.critical(f"Account deleted. Time for me to rest.")
        except KeyboardInterrupt:
            logger.critical("Keyboard Interrupt. Exiting..")
        logger.info("For support visit @{}".format(self.__support__))
        raise SystemExit

    @staticmethod
    def list_modules(directory):
        """List all modules in a directory"""
        if not os.path.exists(directory):
            Stark.log(f"No directory named '{directory}' found")
            return
        if "/." not in directory:
            directory = directory.replace('.', '/')
        plugs = []
        for path, _, files in os.walk(directory):
            for name in files:
                if name.endswith(".py"):
                    sep = "\\" if "\\" in path else "/"
                    data: list = path.split(sep, 1)
                    data.pop(0)
                    if not data:
                        file = name
                    else:
                        file = os.path.join(data[0], name)
                    plugs.append(file[:-3].replace("\\", "/"))
        return plugs

    def load_modules(self, plugins: str):
        """Load all modules from a directory or a single module as pyrogram handlers"""
        sep = "\\" if "\\" in plugins else "/"
        if plugins.endswith(".py"):  # Absolute Path
            # C Drive can't have relative path to D: Drive in Windows so needs absolute path
            plugins, file = plugins.rsplit(sep, 1)
            modules: list[str] = [file[:-3]]
        else:
            modules: list[str] = self.list_modules(plugins)
        plugins = plugins.replace(sep, ".")
        if not modules:
            return
        for module in modules:
            sep = "\\" if "\\" in module else "/"
            module = module.replace(sep, ".")
            if module.endswith(".py"):
                module = module[:-3]
            mod = importlib.import_module(plugins+"."+module)
            funcs = [func for func, _ in getmembers(mod, isfunction)]
            for func in funcs:
                try:
                    for handler, group in getattr(mod, func).handlers:
                        self.add_handler(handler, group)
                except AttributeError:  # Other Functions shouldn't be included
                    pass
            module = module.replace(".", "/")
            __data__["total_plugins"] += 1
            __data__["all_plugins"][module] = mod.__file__
            if "addons" in plugins:
                logger.info("Loaded addon - {}".format(module))
            else:
                logger.info("Loaded plugin - {}".format(module))

    @staticmethod
    def log(message, level: Union[str, int] = logging.INFO):
        """Log messages to console.

        | String       | Integer  |
        |:------------:|:--------:|
        | **debug**    | 10       |
        | **info**     | 20       |
        | **warning**  | 30       |
        | **error**    | 40       |
        | **critical** | 50       |

        Parameters:

            message (Any): Item to print to console.
            level (optional): Logging level as string or int.
        """
        if isinstance(level, str):
            level = level.lower()
        if level in ["critical", 50]:
            level = logging.CRITICAL
        elif level in ["error", 40]:
            level = logging.ERROR
        elif level in ["warning", "warn", 30]:
            level = logging.WARNING
        elif level in ["info", 20]:
            level = logging.INFO
        elif level in ["debug", 10]:
            level = logging.DEBUG
        logger.log(level, message)

    @property
    def data(self) -> dict:
        """A special dictionary with five keys.

        | Key                | Returns                   |
        |:------------------:|:-------------------------:|
        | **total_plugins**  | number of plugins in bot  |
        | **all_plugins**    | dictionary of all plugins and their absolute paths|
        | **total_commands** | number of commands in bot |
        | **all_commands**   | dictionary of command and their descriptions if passed in ``Stark.cmd`` decorator else None |
        | **sudo_commands**  | list of all sudo commands available and includes `owner_only` commands |

        Example:

            ```python
            {
                "total_plugins": 2,
                "all_plugins": {
                    "basic":"C:\\....\\plugins\\basic.py",
                    "sample":"D:\\...\\sample.py"
                },
                "total_commands": 3,
                "all_commands": {
                    "start": "Start the bot"
                    "help": "Get help"
                    "about": "About this bot"
                },
                "sudo_commands": ["stats", "json", "sudo", "botban"]
            }
            ```
        """
        __data__.update(command_data)
        return __data__

    @property
    def total_plugins(self) -> int:
        """Number of total plugins loaded."""
        return self.data["total_plugins"]

    @property
    def all_plugins(self) -> dict:  # not "plugins" to prevent clash with pyrogram's "plugins" attribute
        """Dictionary of all plugins and their respective absolute paths.

        Example:

        ```python
        {
            "basic":"C:\\Users\\....\\plugins\\basic.py",
            "sample":"D:\\Bots\\...\\sample.py",
            "must_join":"C:\\Users\\...\\pystark\\addons\\must_join.py"
        },
        ```
        """
        return self.data["all_plugins"]

    @property
    def total_commands(self) -> int:
        """Number of total commands loaded."""
        return self.data["total_commands"]

    @property
    def all_commands(self) -> dict:
        """Dictionary of all commands and their descriptions if present else None.

        Example:

        ```python
        {
            "start": "Start the bot"
            "help": "Get help"
            "about": "About this bot"
        },
        ```
        """
        return self.data["all_commands"]

    @property
    def sudo_commands(self) -> list[str]:
        """List of all sudo commands available. Includes `owner_only` commands too."""
        return self.data["sudo_commands"]

    def update_bot_menu(self):
        dictionary = self.data["all_commands"]
        commands = []
        for key in dictionary:
            if dictionary[key]:
                commands.append(BotCommand(key, str(dictionary[key])).write())
        self.send(
            raw.functions.bots.SetBotCommands(
                scope=raw.types.BotCommandScopeDefault(),
                lang_code='en',
                commands=commands
            )
        )

    def remove_bot_menu(self):
        self.send(
            raw.functions.bots.ResetBotCommands(
                scope=raw.types.BotCommandScopeDefault(),
                lang_code='en',
            )
        )

    def _load_sql_model(self, name: str):
        tables = [Users, Bans]
        for t in tables:
            if t.__tablename__ == name:
                t.__table__.create(checkfirst=True)
                if name == "users":
                    from pystark.database.sql import Database
                    db = Database()
                    if "lang" not in asyncio.get_event_loop().run_until_complete(db.columns("users")):
                        # Temporary Way
                        asyncio.get_event_loop().run_until_complete(db.add_column("users", "lang", "text"))
                self.log(f"Loaded Database Table: {name}")
                break

    @property
    async def langs(self) -> Union[dict, None]:
        """Returns all languages and corresponding file names if localization is set up otherwise None"""
        return await get_all_langs()

    async def log_tg(self, text):
        """Log a text message to your log chat as defined in environment variable [LOG_CHAT](/start/variables#log_chat)

        Parameters:
            text: Text that needs to be logged.
        """
        await self.send_message(ENV().LOG_CHAT, text)
        # No exceptions are handled for now.

    # def _set_info(self):
    #     bot = self.get_me()
    #     self.__id__ = bot.id
    #     self.__username__ = bot.username
